package dev.fyloz.colorrecipesexplorer.config

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.config.properties.MaterialTypeProperties
import dev.fyloz.colorrecipesexplorer.emergencyMode
import dev.fyloz.colorrecipesexplorer.rest.CRE_PROPERTIES
import dev.fyloz.colorrecipesexplorer.restartApplication
import dev.fyloz.colorrecipesexplorer.service.MaterialTypeService
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationService
import org.slf4j.Logger
import org.springframework.boot.context.event.ApplicationEnvironmentPreparedEvent
import org.springframework.boot.context.event.ApplicationReadyEvent
import org.springframework.context.ApplicationListener
import org.springframework.context.annotation.Configuration
import org.springframework.context.annotation.Profile
import org.springframework.core.Ordered
import org.springframework.core.annotation.Order
import javax.annotation.PostConstruct
import kotlin.concurrent.thread

@Configuration
@Order(Ordered.HIGHEST_PRECEDENCE)
@Profile("!emergency")
class ApplicationReadyListener(
    private val materialTypeService: MaterialTypeService,
    private val configurationService: ConfigurationService,
    private val materialTypeProperties: MaterialTypeProperties,
    private val creProperties: CreProperties,
    private val logger: Logger
) : ApplicationListener<ApplicationReadyEvent> {
    override fun onApplicationEvent(event: ApplicationReadyEvent) {
        if (emergencyMode) {
            logger.error("Emergency mode is enabled, default material types will not be created")
            thread {
                Thread.sleep(1000)
                logger.warn("Restarting in emergency mode...")
                restartApplication(true)
            }
            return
        }

        initDatabaseConfigurations()
        initMaterialTypes()
        CRE_PROPERTIES = creProperties
    }

    private fun initMaterialTypes() {
        logger.info("Initializing system material types")
        materialTypeService.saveSystemTypes(materialTypeProperties.systemTypes)
    }

    private fun initDatabaseConfigurations() {
        configurationService.initializeProperties { !it.file }
    }
}

@Configuration("configurationsInitializer")
class ConfigurationsInitializer(
    private val configurationService: ConfigurationService
) {
    @PostConstruct
    fun initializeFileConfigurations() {
        configurationService.initializeProperties { it.file }
    }
}

class ApplicationInitializer : ApplicationListener<ApplicationEnvironmentPreparedEvent> {
    override fun onApplicationEvent(event: ApplicationEnvironmentPreparedEvent) {
        if (emergencyMode) {
            event.environment.setActiveProfiles("emergency")
        }
    }
}
