package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.service.config.CONFIGURATION_FORMATTED_LIST_DELIMITER
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationServiceImpl
import dev.fyloz.colorrecipesexplorer.service.config.ConfigurationSource
import dev.fyloz.colorrecipesexplorer.utils.encrypt
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import kotlin.UnsupportedOperationException
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class ConfigurationServiceTest {
    private val fileService = mockk<FileService>()
    private val configurationSource = mockk<ConfigurationSource>()
    private val securityProperties = mockk<CreSecurityProperties> {
        every { configSalt } returns "d32270943af7e1cc"
    }
    private val service = spyk(ConfigurationServiceImpl(fileService, configurationSource, securityProperties, mockk()))

    @AfterEach
    fun afterEach() {
        clearAllMocks()
    }

    // getAll()

    @Test
    fun `getAll() gets the Configuration of each ConfigurationType`() {
        every { service.get(any<ConfigurationType>()) } answers { throw ConfigurationNotSetException(this.args[0] as ConfigurationType) }

        service.getAll()

        verify {
            service.getAll()
            ConfigurationType.values().forEach {
                service.get(it)
            }
        }
        confirmVerified(service)
    }

    @Test
    fun `getAll() only returns set configurations`() {
        val unsetConfigurationTypes = listOf(
            ConfigurationType.INSTANCE_NAME,
            ConfigurationType.INSTANCE_LOGO_PATH,
            ConfigurationType.INSTANCE_ICON_PATH
        )

        every { service.get(match<ConfigurationType> { it in unsetConfigurationTypes }) } answers {
            throw ConfigurationNotSetException(this.firstArg() as ConfigurationType)
        }
        every { service.get(match<ConfigurationType> { it !in unsetConfigurationTypes }) } answers {
            val type = firstArg<ConfigurationType>()
            configuration(type, type.key)
        }

        val found = service.getAll()

        assertFalse {
            found.any {
                it.type in unsetConfigurationTypes
            }
        }

        verify {
            service.getAll()
            ConfigurationType.values().forEach {
                service.get(it)
            }
        }
        confirmVerified(service)
    }

    @Test
    fun `getAll() only includes configurations matching the formatted formatted key list`() {
        val configurationTypes = listOf(
            ConfigurationType.INSTANCE_NAME,
            ConfigurationType.INSTANCE_LOGO_PATH,
            ConfigurationType.INSTANCE_ICON_PATH
        )
        val formattedKeyList = configurationTypes
            .map { it.key }
            .reduce { acc, s -> "$acc$CONFIGURATION_FORMATTED_LIST_DELIMITER$s" }

        every { service.get(any<String>()) } answers {
            val key = firstArg<String>()
            configuration(key.toConfigurationType(), key)
        }

        val found = service.getAll(formattedKeyList)

        assertTrue {
            found.all { it.type in configurationTypes }
        }

        verify {
            service.getAll(formattedKeyList)
            configurationTypes.forEach {
                service.get(it.key)
            }
        }
        confirmVerified(service)
    }

    // get()

    @Test
    fun `get(key) calls get() with the ConfigurationType matching the given key`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH
        val key = type.key

        every { service.get(type) } answers {
            val type = firstArg<ConfigurationType>()
            configuration(type, type.key)
        }

        service.get(key)

        verify {
            service.get(key)
            service.get(type)
        }
        confirmVerified(service)
    }

    @Test
    fun `get(type) gets the configuration in the ConfigurationSource`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH
        val configuration = configuration(type = type)

        every { configurationSource.get(type) } returns configuration

        val found = service.get(type)

        assertEquals(configuration, found)
    }

    @Test
    fun `get(type) throws ConfigurationNotSetException when the given ConfigurationType has no set configuration`() {
        val type = ConfigurationType.INSTANCE_ICON_PATH

        every { configurationSource.get(type) } returns null

        with(assertThrows<ConfigurationNotSetException> { service.get(type) }) {
            assertEquals(type, this.type)
        }

        verify {
            service.get(type)
            configurationSource.get(type)
        }
    }

    @Test
    fun `get(type) throws InvalidConfigurationKeyException when the given ConfigurationType is encryption salt`() {
        val type = ConfigurationType.GENERATED_ENCRYPTION_SALT

        assertThrows<InvalidConfigurationKeyException> { service.get(type) }
    }

    @Test
    fun `get(type) returns a SecureConfiguration when the given ConfigurationType is secure`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val configuration = configuration(
            type = type,
            content = "securepassword".encrypt(type.key, securityProperties.configSalt!!)
        )

        every { configurationSource.get(type) } returns configuration

        val found = service.get(type)

        assertTrue { found is SecureConfiguration }
    }

    @Test
    fun `getContent(type) returns configuration content`() {
        val type = ConfigurationType.INSTANCE_NAME
        val configuration = configuration(
            type = type,
            content = "content"
        )

        every { service.get(type) } returns configuration

        val found = service.getContent(type)

        assertEquals(configuration.content, found)
    }

    @Test
    fun `getContent(type) throws UnsupportedOperationException when configuration is secure`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val configuration = secureConfiguration(type)

        every { service.get(type) } returns configuration

        assertThrows<UnsupportedOperationException> { service.getContent(type) }
    }

    @Test
    fun `getSecure(type) returns decrypted configuration content`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val content = "securepassword"
        val configuration = configuration(
            type = type,
            content = content.encrypt(type.key, securityProperties.configSalt!!)
        )

        every { configurationSource.get(type) } returns configuration

        val found = service.getSecure(type)

        assertEquals(content, found)
    }

    @Test
    fun `getSecure(type) throws UnsupportedOperationException when configuration is not secure`() {
        val type = ConfigurationType.INSTANCE_NAME

        assertThrows<UnsupportedOperationException> { service.getSecure(type) }
    }

    @Test
    fun `set(configuration) set configuration in ConfigurationSource`() {
        val configuration = configuration(type = ConfigurationType.INSTANCE_NAME)

        every { configurationSource.set(any<Configuration>()) } just runs

        service.set(configuration)

        verify {
            configurationSource.set(configuration)
        }
    }

    @Test
    fun `set(configuration) encrypts secure configurations`() {
        val type = ConfigurationType.DATABASE_PASSWORD
        val content = "securepassword"
        val encryptedContent = content.encrypt(type.key, securityProperties.configSalt!!)
        val configuration = configuration(type = type, content = content)

        mockkStatic(String::encrypt)

        every { configurationSource.set(any<Configuration>()) } just runs
        every { content.encrypt(any(), any()) } returns encryptedContent

        service.set(configuration)

        verify {
            configurationSource.set(match<Configuration> {
                it.content == encryptedContent
            })
        }
    }
}
